/* Ovladani hlasem: Borceni casove osy */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <dirent.h>

#define ISQRT2 0.707106781186547524400844362104849
#define PI 3.1415926535897932384626433832795

#define ERROR(message) (printf message, exit(-1))

/* Realna jednorozmerna in-place FFT: Sorensenuv split-radix algoritmus
 * [Sorensen et al: Real-Valued Fast Fourier Transform Algorithms, IEEE Trans. ASSP, ASSP-35, No. 6, June 1987]
 * in: re(0),re(1),re(2),...,re(size-1)
 * out (s normalizaci): re(0),re(1),re(2),...,re(size/2),im(size/2-1),...,im(1)
 */
void rfft(float* data, int n)
{ int i,j,k,i5,i6,i7,i8,i0,id,i1,i2,i3,i4,n2,n4,n8;
  float t1,t2,t3,t4,t5,t6,a,e;
  float invn = 1.0 / n;

  n4 = n-1;

  // data shuffling
  for (i = 0, j = 0, n2 = n/2; i < n4; j += k, i++)
  { if (i < j) t1 = data[j], data[j] = data[i], data[i] = t1;
    for (k = n2; k <= j; k >>= 1) j -= k;
  }

  // length two butterflies
  i0 = 0;
  id = 4;
  do
  { for ( ; i0 < n4; i0 += id)
    { i1 = i0+1;
      t1 = data[i0];
      data[i0] = t1+data[i1];
      data[i1] = t1-data[i1];
    }
    id <<= 1;
    i0 = id-2;
    id <<= 1;
  } while (i0 < n4);

  // L shaped butterflies
  n2 = 2;
  for (k = n; k > 2; k >>= 1)
  { n2 <<= 1;
    n4 = n2>>2;
    n8 = n2>>3;
    e = 2*PI/n2;
    i1 = 0;
    id = n2<<1;
    do
    { for ( ; i1 < n; i1 += id)
      { i2 = i1+n4; i3 = i2+n4; i4 = i3+n4;
        t1 = data[i4]+data[i3];
        data[i4] -= data[i3];
        data[i3] = data[i1]-t1;
        data[i1] += t1;
        if (n4 != 1)
        { i0 = i1+n8; i2 += n8; i3 += n8; i4 += n8;
          t1 = (data[i3]+data[i4]) * ISQRT2;
          t2 = (data[i3]-data[i4]) * ISQRT2;
          data[i4] = data[i2]-t1;
          data[i3] = -data[i2]-t1;
          data[i2] = data[i0]-t2;
          data[i0] += t2;
      } }
      id <<= 1;
      i1 = id-n2;
      id <<= 1;
    } while (i1 < n);
    a = e;
    for (j = 2; j <= n8; j++)
    { float ss1 = sin(a), cc1 = cos(a), ss3 = sin(3*a), cc3 = cos(3*a);
      a = j*e;
      i = 0;
      id = n2 << 1;
      do
      { for ( ; i < n ; i += id)
        { i1 = i+j-1; i2 = i1+n4; i3 = i2+n4; i4 = i3+n4; i5 = i+n4-j+1; i6 = i5+n4; i7 = i6+n4; i8 = i7+n4;
          t1 = data[i3]*cc1 + data[i7]*ss1;
          t2 = data[i7]*cc1 - data[i3]*ss1;
          t3 = data[i4]*cc3 + data[i8]*ss3;
          t4 = data[i8]*cc3 - data[i4]*ss3;
          t5 = t1+t3;
          t3 = t1-t3;
          t6 = t2+t4;
          t4 = t2-t4;
          t2 = data[i6]+t6; data[i3] = -data[i6]+t6; data[i8] = t2;
          t2 = data[i2]-t3; data[i7] = -data[i2]-t3; data[i4] = t2;
          t1 = data[i1]+t5; data[i6] = data[i1]-t5; data[i1] = t1;
          t1 = data[i5]+t4; data[i5] -= t4; data[i2] = t1;
        }
        id <<= 1;
        i = id-n2;
        id <<= 1;
      } while (i < n);
  } }

  // division with array length
  for (i = 0; i < n; i++) data[i] *= invn;
}


float mel2f(float mel) { return (pow(10, mel/2595) - 1) * 700; }
float f2mel(float f) { return 2595 * log10(1 + f/700); }


/* Mel-Frequency Cepstral Coefficients; meni vstupni data
 * Parametry:
 *  rate       = vzorku za sekundu
 *  fft_size   = vzorku v okne
 *  minf, maxf = hranice frekvence
 *  filters    = pocet filtru
 *  mfcc_count = pocet koeficientu v sade
 */
void mfcc(float *a, float *dst, float rate, int fft_size, float minf, float maxf, int filters, int mfcc_count)
{ int i, j;
  float m[filters];

  // copy, Hamming
  for (i = 0; i < fft_size; i++)
    a[i] *= 0.54 - 0.46 * cos(i * PI / fft_size);

  // FFT magnitude
  rfft(a, fft_size);
  for (i = 1; i < fft_size/2; i++)
    a[i] = sqrt(a[i] * a[i] + a[fft_size-i] * a[fft_size-i]);

  // Mel filterbank
  float mmin = f2mel(minf),
        mmax = f2mel(maxf),
        mstep = (mmax - mmin) / (filters + 2);

  for (i = 0; i < filters; i++)
  { float w;

    int j0 = mel2f(mmin + i*mstep) * fft_size/rate,
        j1 = mel2f(mmin + (i+1)*mstep) * fft_size/rate,
        j2 = mel2f(mmin + (i+2)*mstep) * fft_size/rate;
    for (j = j0, m[i] = w = 0; j < j1; j++, w += 1./(j1 - j0))
      m[i] += w * a[j];
    for (j = j1, w = 1; j < j2; j++, w -= 1./(j2 - j1))
      m[i] += w * a[j];

    m[i] = log(m[i] / (j2-j0));
  }

  // DCT
  for (i = 0; i < mfcc_count; i++)
    for (j = 0, dst[i] = 0; j < filters; j++)
      dst[i] += m[j] * cos((i+1) * PI / filters * (j-0.5));
}



/* Dynamic Time Warping
 * pro kazde policko spocitam podobnost p(i,j) a smer, odkud prijit
 * d(i,j) = p(i,j) + min(d(i-1, j-1), d(i-1, j), d(i, j-1))
 */
float dtw(float* a, int alen, float* b, int blen, int coeffs)
{ int i, j, k;
  float D[2][blen];  // pamatuji si minulou radku
  float* d = &(D[0][0]);
  float* lastd = &(D[1][0]);

  for (i = 0; i < alen; i++)
  { for (j = 0; j < blen; j++)
    { for (k = 0, d[j] = 0; k < coeffs; k++)
        d[j] += (a[coeffs*i + k] - b[coeffs*j + k]) * (a[coeffs*i + k] - b[coeffs*j + k]);
      d[j] = sqrt(d[j]);

      float min = 1e30;
      if (j && d[j-1] < min) min = d[j-1];
      if (i && lastd[j] < min) min = lastd[j];
      if (i && j && lastd[j-1] < min) min = lastd[j-1];
      if (i || j) d[j] += min;

    }
    float* tmp = d;  // prohodime ukazatele
    d = lastd;
    lastd = tmp;
  }
  return lastd[blen-1];
}


// vrati delku souboru
int flength(char *file)
{ FILE *f;
  if (NULL == (f = fopen(file, "rb"))) ERROR(("Soubor nejde otevrit\n"));
  fseek(f, -1, SEEK_END);
  int len = ftell(f) + 1;
  fclose(f);
  return len;
}

/*
 * dostaneme 11025 Hz zvuk v int16_t
 * spocitame jeho MFCC koeficienty v case
 * pomoci DTW je porovname se vsemi sablonami ze souboru
 */
int main(int argc, char** argv)
{ int i, j;

  if (!(argc>2 && argv[1][0]=='-' && ( (argv[1][1]=='c' && argc==4) || (argv[1][1]=='d' && argc==3) ) ))
    ERROR(("EasyDTW\n\nPouziti:\n Prevod zvuku na MFCC: easydtw -c zvuk.raw zvuk.mfc\n DTW vuci *.mfc: easydtw -d zvuk.mfc\n"));

  switch (argv[1][1])
  { case 'c':
    { FILE* fi;
      FILE* fo;
      int copied;
      signed short wave[512];
      float a[512];
      float m[12];

      if (NULL == (fi = fopen(argv[2], "rb"))) ERROR(("Soubor nejde otevrit\n"));
      if (NULL == (fo = fopen(argv[3], "wb"))) ERROR(("Soubor nejde otevrit\n"));

      int pointer = 0;
      for (;;)
      { fseek(fi, 256*pointer*sizeof(short), SEEK_SET); // overlap = 256
        if (!(copied = fread(wave, 2, 512, fi))) break; // fft_size = 512
        for (i = 0; i < copied; i++) a[i] = wave[i];    // konverze do floatu
        for (i = copied; i < 512; i++) a[i] = 0;        // a doplnime nulami
        mfcc(a, m, 11025, 512, 130, 5400, 36, 12);
        fwrite(m, 4, 12, fo);
        pointer++;
      }
      fclose(fi);
      fclose(fo);
      break;
    }

    case 'd':
    { DIR* d = opendir(".");
      struct dirent *de;
      float* a;
      float* b;
      int alen, blen;
      FILE* f;

      alen = flength(argv[2]) / (12*sizeof(float));
      f = fopen(argv[2], "rb");
      a = malloc(alen * (12*sizeof(float)));
      fread(a, 12*sizeof(float), alen, f);
      fclose(f);

      float actual, lowest = 1e30;
      char lowest_file[260];
      while (de = readdir(d))
      { if (strlen(de->d_name) >= 4 && strcmp(de->d_name + strlen(de->d_name) - 4, ".mfc") == 0)
        { blen = flength(de->d_name) / (12*sizeof(float));
          f = fopen(de->d_name, "rb");
          b = malloc(blen * (12*sizeof(float)));
          fread(b, 12*sizeof(float), blen, f);
          fclose(f);
          printf("\n%s: %f", de->d_name, actual = dtw(a, alen, b, blen, 12));
          free(b);
          if (actual > 1 && actual < lowest) lowest = actual, strcpy(lowest_file, de->d_name);
        }
      }
      closedir(d);
      free(a);

      printf("\nBest match: %s (%f)\n", lowest_file, lowest);
      break;
  } }
  return 0;
}
